/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.canvas;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.fmap.core.IGeometry;
import es.prodevelop.gvsig.mobile.fmap.symbol.FSymbol;
import es.prodevelop.gvsig.mobile.fmap.util.graph2d.Graph2DUtilities;
import es.prodevelop.gvsig.mobile.fmap.viewport.ViewPort;

/**
 * This is a second light canvas where measurement geometries (distance, area)
 * are drawn.
 * 
 * @see es.prodevelop.gvsig.mobile.gui.canvas.MapControl
 * @see es.prodevelop.gvsig.mobile.fmap.viewport.ViewPort
 * @see es.prodevelop.gvsig.mobile.gui.tool.listener.DistanceAreaListener
 * 
 *  
 * @author jldominguez
 *
 */
public class TransparentCanvas extends Component {
	
	private static Logger logger = Logger.getLogger(TransparentCanvas.class);
	
	private ViewPort viewPort;
	private ArrayList geometryList;
	private ArrayList geometryEditionList;
	private static TransparentCanvas trans = null;

	public static TransparentCanvas getInstance(){
		if(trans == null){
			trans = new TransparentCanvas();
		}
		return trans;
	}


	/**
	 * Constructor.
	 * 
	 * @param vp A reference to the viewport.
	 */
	private TransparentCanvas() {

		geometryList = new ArrayList();
		geometryEditionList =  new ArrayList();
	}
	
	/**
	 * Sets  a new viewport 
	 * @param vp the new viewport
	 */
	public void setViewPort(ViewPort vp) {
		viewPort = vp;
	}
	
	/**
	 * Gets the viewport
	 * @return the viewport
	 */
	public ViewPort getViewPort() {
		return viewPort;
	}
	
	/**
	 * Adds a new geometry to this canvas. Normally there is only one or zero geometries,
	 * (it is constantly being cleared).
	 * 
	 * @param ig the new geometry to be added
	 * @return number of geometries currently in this canvas (after action)
	 */
	public int addGeometry(IGeometry ig) {
		geometryList.add(ig);
		return geometryList.size(); 
	}

	public int addEditionGeometry(IGeometry ig){
		geometryEditionList.add(ig);
		return geometryEditionList.size();
	}

	/**
	 * Removes a geometry from the current canvas
	 * @param ig the geometry to be removed
	 * @return number of geometries currently in this canvas (after action)
	 */
	public int removeGeometry(IGeometry ig) {
		geometryList.remove(ig);
		return geometryList.size(); 
	}

	public int removeEditionGeometry(IGeometry ig){
		geometryEditionList.remove(ig);
		return geometryEditionList.size();
	}

	/**
	 * overridden pain method
	 */
	public void paint(Graphics _g) {
		paint(_g, true, FSymbol.MEASURE_SYMBOL);
	}
	
	/**
	 * Paint method.
	 * 
	 * @param _g the graphics object where geometries are going to be drawn
	 * 
	 * @param transf whether or not the affine transformation of the viewport
	 * has to be applied to geometries before drawing.
	 * 
	 * @param sym symbol to be used for painting geometries
	 */
	public void paint(Graphics _g, boolean transf, FSymbol sym) {
		
		Graphics2D g = (Graphics2D) _g; 
		AffineTransform at = viewPort.getAffineTransform();

		IGeometry ig = null;
		Shape shp = null;
		
		_g.setColor(sym.getBorderColor());

		for (int i=0; i<geometryList.size(); i++) {
			
			ig = (IGeometry) geometryList.get(i);

			if (ig != null) {
				ig = ig.cloneGeometry();
				if (transf) ig.transform(at);
				shp =  ig.getInternalShape();
				Graph2DUtilities.drawShape(g, shp, sym);
			}
		}

		for (int i=0; i<geometryEditionList.size(); i++) {

			ig = (IGeometry) geometryEditionList.get(i);

			if (ig != null) {
				ig = ig.cloneGeometry();
				if (transf) ig.transform(at);
				shp =  ig.getInternalShape();
				Graph2DUtilities.drawShape(g, shp, sym);
			}
		}
	}
	
	/**
	 * Gets the i-th geometry in the canvas.
	 * 
	 * @param i index of the requested geometry
	 * @return the requested geometry or null if index >= size
	 */
	public IGeometry getGeometry(int i) {
		if (geometryList.size() == 0) return null;
		return (IGeometry) geometryList.get(i);
	}

	public IGeometry getEditionGeometry(int i) {
		if (geometryEditionList.size() == 0) return null;
		return (IGeometry) geometryEditionList.get(i);
	}

	/**
	 * Removes all geometries.
	 * 
	 * @return whether any geometry was removed. Returns false if canvas was already empty
	 */
	public boolean clear() {
		if (geometryList.size() == 0) return false;
		
		geometryList.clear();
		return true;
	}

	public boolean clearEdition(){
		if(geometryEditionList.size() == 0) return false;

		geometryEditionList.clear();
		return true;
	}

}
