/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.dialog.control;

import java.awt.Color;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.fmap.layer.FLayer;
import es.prodevelop.gvsig.mobile.fmap.layer.FLyrVect;
import es.prodevelop.gvsig.mobile.fmap.legend.IntervalLegend;
import es.prodevelop.gvsig.mobile.fmap.legend.SingleSymbolLegend;
import es.prodevelop.gvsig.mobile.fmap.symbol.FSymbol;

/**
 * This class keeps a copy of the symbology and other TOC settings and is used
 * to detect changes in the symbollogy that must cause a redrawing of the map
 * and a refresh of the symbology icons in the TOC.
 *  
 *  
 *  @see es.prodevelop.gvsig.mobile.gui.dialog.control.LayersTableItem
 *  
 * @author jldominguez
 *
 */
public class ViewLayersGraphicStatus {
	
	private static Logger logger = Logger.getLogger(ViewLayersGraphicStatus.class);
	
	private int count = 0;
	private boolean[] visible;
	private boolean[] filled;
	private long[] layerids;
	private Color[] borderColors;
	private Color[] fillColors;
	private int[] strokeWidths;
	private double[] maxScales;
	private double[] minScales;
	private boolean[] isInterval;
	private IntervalLegend[] legends;
	
	/**
	 * Gets the ID of the i-th layer considered in this object
	 * @param i index of the referenced layer 
	 * @return ID of the referenced layer
	 */
	public long getId(int i) {
		if (i >= count) return 0;
		return layerids[i];
	}
	
	/**
	 * Gets the max scale of the i-th layer in this object
	 * @param i the index of the layer 
	 * @return the max scale of the i-th layer in this object
	 */
	public double getMaxScale(int i) {
		if (i >= count) return -1;
		return maxScales[i];
	}

	/**
	 * Gets the min scale of the i-th layer in this object
	 * @param i the index of the layer 
	 * @return the min scale of the i-th layer in this object
	 */
	public double getMinScale(int i) {
		if (i >= count) return -1;
		return minScales[i];
	}

	/**
	 * Gets whether the i-th layer in this object is visible
	 * @param i the index of the layer 
	 * @return true if the i-th layer in this object is visible, false otherwise
	 */
	public boolean getVisible(int i) {
		if (i >= count) return false;
		return visible[i];
	}
	
	/**
	 * Gets the border color used in the symbology of the i-th layer
	 * @param i the index of the layer 
	 * @return the border color used in the symbology of the layer
	 */
	public Color getBorderColor(int i) {
		if (i >= count) return null;
		return borderColors[i];
	}

	/**
	 * Gets the fill color used in the symbology of the i-th layer
	 * @param i the index of the layer 
	 * @return the fill color used in the symbology of the layer
	 */
	public Color getFillColor(int i) {
		if (i >= count) return null;
		return fillColors[i];
	}
	
	/**
	 * Gets whether the symbology of i-th layer is filled for polygons
	 * @param i the index of the layer 
	 * @return whether the symbology of i-th layer is filled for polygons
	 */
	public boolean getFilled(int i) {
		if (i >= count) return false;
		return filled[i];
	}
	
	
	/**
	 * Gets the width of the stroke used in the symbology of the i-th layer
	 * @param i the idnex of the layer
	 * @return the width of the stroke used in the symbology of the i-th layer
	 */
	public int getStrokeWidth(int i) {
		if (i >= count) return 0;
		return strokeWidths[i];
	}
	
	/**
	 * Returns true if the layer i uses interval legend
	 * @param i
	 * @return
	 */
	public boolean getUseInterval(int i) {
		if (i >= count) return false;
		return this.isInterval[i];
	}	
	
	private IntervalLegend getIntervalLegend(int i) {
		if (i >= count) return null;
		return this.legends[i];
	}

	

	/**
	 * Constructor
	 * 
	 * @param t The LayersTable to be monitored
	 * 
	 */
	public ViewLayersGraphicStatus(LayersTable t) {
		FLayer[] layers = t.getLayers();
		
		count = layers.length;
		layerids = new long[count];
		visible = new boolean[count];
		filled = new boolean[count];
		borderColors = new Color[count];
		fillColors = new Color[count];
		strokeWidths = new int[count];
		
		isInterval = new boolean[count];
		
		maxScales = new double[count];
		minScales = new double[count];
		
		legends = new IntervalLegend[count];

		for (int i=0; i<count; i++) {
			visible[i] = layers[i].isVisible();
			layerids[i] = layers[i].getId();
			
			maxScales[i] = layers[i].getMaxScale();
			minScales[i] = layers[i].getMinScale();
			
			if (layers[i] instanceof FLyrVect) {
				
				FLyrVect lyr = (FLyrVect) layers[i];
				SingleSymbolLegend leg = (SingleSymbolLegend) lyr.getLegend();
				FSymbol sym = leg.getDefaultSymbol();
				borderColors[i] = sym.getBorderColor();
				fillColors[i] = sym.getFillColor();
				filled[i] = sym.isFilled();
				strokeWidths[i] = sym.getStrokeWidth();
				isInterval[i] = lyr.IsUsedIntervalLegend();
				if (lyr.getIntervalLegend() != null)
					legends[i] = lyr.getIntervalLegend().cloneIntervalLegend();

			} else {

				borderColors[i] = Color.BLACK;
				filled[i] = false;
				fillColors[i] = Color.BLACK;
				strokeWidths[i] = -1;
				
			}
		} 
	}
	
	public int getLayerCount() {
		return count;
	}
	
	/**
	 * Tells whether there have been changes between the old TOC settings
	 * and the ones provided
	 * 
	 * @param stt TOC settings to be compared with previous ones
	 * 
	 * @return whether there have been changesin the TOC settings
	 */
	public boolean isSame(ViewLayersGraphicStatus stt) {
		
		int size = stt.getLayerCount();
		if (count != size) return false;
		
		for (int i=0; i<size; i++) {
			if (stt.getId(i) != getId(i)) return false;
			
			if (stt.getMinScale(i) != getMinScale(i)) return false;
			if (stt.getMaxScale(i) != getMaxScale(i)) return false;

			if (stt.getVisible(i) != getVisible(i)) return false;
			if (stt.getStrokeWidth(i) != getStrokeWidth(i)) return false;
			if (stt.getFilled(i) != getFilled(i)) return false;
			if (stt.getFillColor(i).getRGB() != getFillColor(i).getRGB()) return false;
			if (stt.getBorderColor(i).getRGB() != getBorderColor(i).getRGB()) return false;
			if (stt.getUseInterval(i) != getUseInterval(i)) return false;
			IntervalLegend l = stt.getIntervalLegend(i);
			if (l != null)
				if (!l.isSame(getIntervalLegend(i))) return false;
		}
		return true;
	}
	

}
