/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.tab;

import java.awt.Component;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.gui.dialog.DialogConstants;
import es.prodevelop.gvsig.mobile.gui.dialog.control.ExpansibleRow;

/**
 * 
 * This class is used to create panels expandible horizontally
 * and vertically.
 * 
 * Each row of the panel will have at most one control
 * that will be added in the expandible way (horizonatal expansion).
 * 
 * Each <code>ExpansiblePanel</code> will have at most one row
 * that is added in the expansible way (vertical expansion)
 * 
 * @see es.prodevelop.gvsig.mobile.gui.dialog.control.ExpansibleRow
 * 
 * @author jcarrasco
 *
 */
public class ExpansiblePanel extends Panel {
	
	private static Logger logger = Logger.getLogger(ExpansiblePanel.class);
	
	private ExpansibleRow extCom;
	
	/**
	 * Constructor with default initial size
	 *
	 */
	public ExpansiblePanel() {
		super();
		setSize(100, 20);
		initialize();
	}
	
	/**
	 * Constructor with specific initial size
	 * @param w initial panel width
	 * @param h initial panel height
	 */
	public ExpansiblePanel(int w, int h) {
		super();
		setSize(w, h);
		initialize();
	}
	
	/**
	 * Sets panel size
	 * 
	 * @param w new panel width
	 * @param h new panel height
	 */
	public void setSize(int w, int h) {
		super.setSize(w, h);
	}

	private void initialize() {
		setLayout(null);

		addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				expand();
			}
		});
		
	}

	/**
	 * This method is called when all components have been added or updated
	 * and the panel must adopt a sensible size.
	 *
	 */
	public void expand() {
		//If the panel height is bigger than the constrols total height
		//the  expansiblerow will grow in height to complete the panel
		//height, else, the panel will grow in height to fit the controls
		Component[] coms = getComponents();
		
		if (extCom!=null) {
			int newHeight = getMaxHeight();
			if (newHeight >= extCom.getMinHeight()){
				extCom.setSize(getWidth(), newHeight);
				extCom.growHeight();
			} else {
				//This panel is smaller than its content
				//we can't grow any control, we should grow
				//the panel
				this.setSize(this.getWidth(), this.getHeight() + extCom.getMinHeight() - newHeight);
			}
		}
		
		int acy = DialogConstants.YSTART;
		for (int i=0; i< coms.length; i++){
			if (! (coms[i] instanceof ExpansibleRow))
				continue;
			ExpansibleRow row = (ExpansibleRow) coms[i];
			row.setLocation(DialogConstants.XSPACE, acy);
			row.expand(this.getWidth());
			acy += coms[i].getHeight() + DialogConstants.YSPACE;
		}
	}
	
	/**
	 * Adds a new component the panel. It will be treated as a
	 * new non-vertically expansible row
	 * 
	 *  @param com the new component to be added (treated as a row)
	 */
	public Component add(Component com){
		Component resp = super.add(com);
		return resp;
	}
	
	/**
	 * Adds a new component the panel. It will be treated as a
	 * new vertically expansible row (there must be zero or one
	 * vertically expansible rows in the panel)
	 * 
	 *  @param row the new component to be added (treated as a row)
	 */
	public void addExpansible(ExpansibleRow row){
		extCom = row;
		add(row);
	}
	
	/**
	 * Gets the maximum height that the panel can have with the current components.
	 * @return the maximum height that the panel can have with the current components
	 */
	public int getMaxHeight() {

		Component[] coms = getComponents();
		int ach = DialogConstants.YSTART;
		for (int i=0; i< coms.length; i++){
			ach += coms[i].getHeight();
			ach += DialogConstants.YSPACE;
		}
		ach += DialogConstants.YSPACE;

		if (extCom != null) {
			int resp = getHeight() - ach + extCom.getHeight();
			return resp;
		}
		return getHeight() - ach;
	}
	
	public void actionPerformed(ActionEvent e) { }

}
