/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.tool.listener;

import java.awt.Cursor;
import java.awt.Frame;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.fmap.core.GeneralPathX;
import es.prodevelop.gvsig.mobile.fmap.core.IGeometry;
import es.prodevelop.gvsig.mobile.fmap.core.ShapeFactory;
import es.prodevelop.gvsig.mobile.gui.canvas.MapControl;
import es.prodevelop.gvsig.mobile.gui.tool.event.PointEvent;
import es.prodevelop.gvsig.mobile.gui.tool.listener.imterfaces.PointToolListener;

/**
 * This listener is associated with the measurement tools (distance/area).
 * It will compute areas / accumulative distances each time a point event happens.
 * 
 * Of course it needs geographic information about scales, provided by the
 * <code>MapControl</code>.
 * 
 * @see es.prodevelop.gvsig.mobile.gui.tool.behavior.PointBehavior
 * 
 * @author jldominguez
 *
 */
public class DistanceAreaListener implements PointToolListener {
	
	private static Logger logger = Logger.getLogger(DistanceAreaListener.class);
	private MapControl mapCtrl;
	private boolean emptyGpx = true;
	private GeneralPathX gpx = new GeneralPathX();
	
	private int refresh_minx = 0;
	private int refresh_miny = 0;
	private int refresh_maxx = 0;
	private int refresh_maxy = 0;
	
	private boolean isArea = false;
	
	public static int DRAG_LIMIT = 3;

	/**
	 * Constructor. The coordinates of the initial point are needed for refresh
	 * purposes.
	 * 
	 * @param x the x coordinate of the starting point clicked by the user
	 * @param y the y coordinate of the starting point clicked by the user
	 * @param area whether an area (true) or a distance (false) has to be computed.
	 */
	public DistanceAreaListener(int x, int y, boolean area) {
		refresh_minx = x;
		refresh_miny = y;
		isArea = area;
	}
	
	private DistanceAreaListener() {
	}
	
	private void addPoint(Point2D p) {

		IGeometry ig = null;
		if (emptyGpx) {
			emptyGpx = false;
			gpx.moveTo(p.getX(), p.getY());
			ig = ShapeFactory.createPoint2D(p.getX(), p.getY());
		} else {
			gpx.lineTo(p.getX(), p.getY());
			if (isArea) {
				GeneralPathX _gpx = (GeneralPathX) gpx.clone();
				_gpx.closePath();
				ig = ShapeFactory.createPolygon2D(_gpx);
			} else {
				ig = ShapeFactory.createPolyline2D(gpx);
			}
		}
		mapCtrl.setDistanceGeometry(ig, refresh_minx, refresh_miny, refresh_maxx, refresh_maxy);
	}
	
	private void newPointIndicated(int x, int y) {
		updateRefreshLimits(x, y);
		Point2D mp = mapCtrl.getViewPort().toMapPoint(x, y);
		addPoint(mp);
	}

	private void updateRefreshLimits(int x, int y) {
		refresh_maxx = Math.max(refresh_maxx, x) + 1;
		refresh_maxy = Math.max(refresh_maxy, y) + 1;
		refresh_minx = Math.min(refresh_minx, x) - 1;
		refresh_miny = Math.min(refresh_miny, y) - 1;
	}

	/**
	 * This method will be called when a point event happens.
	 * The listener will computer again the area/distance.
	 */
	public void point(PointEvent event) {
		
		int id = event.getEvent().getID();
		switch (id) {
		
		case MouseEvent.MOUSE_PRESSED:
			newPointIndicated(event.getPoint().x, event.getPoint().y);
			updateMeasure();
			break;
		}
	}

	private void updateMeasure() {
		if (isArea) {
			mapCtrl.setSymbolGeometryAreaInStatusBar();
		} else {
			mapCtrl.setSymbolGeometryLengthAsDistanceInStatusBar();
		}
	}

	/**
	 * Gets the cursor to be used while this tool is active.
	 * 
	 * @return the cursor to be used while this tool is active.
	 */
	public Cursor getCursor() {
		// TODO Auto-generated method stub
		return null;
	}

	/**
	 * This method asks the listener to cancel the current drawing process
	 * (currently unsued)
	 * 
	 * @return whether it was possible to cancel the drawing
	 */
	public boolean cancelDrawing() {
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * Sets the <code>MapControl</code> object that will be updated by
	 * this listener.
	 */
	public void setMapControl(MapControl mc) {
		mapCtrl = mc;
	}

	/**
	 * This method cleans the listener (resets distance/area to zer and removes
	 * geometry). It's used when the tool is activated/deactivated.
	 *
	 */
	public void start() {
		emptyGpx = true;
		gpx = new GeneralPathX();
	}
	
}
