/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.tool.listener;

import java.awt.Cursor;
import java.awt.Image;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.util.Timer;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.gui.canvas.MapControl;
import es.prodevelop.gvsig.mobile.gui.tool.event.PointEvent;
import es.prodevelop.gvsig.mobile.gui.tool.listener.imterfaces.PointToolListener;
import es.prodevelop.gvsig.mobile.gui.tool.util.PanRefreshTimerTask;

/**
 * This listener is associated with the pan tool.
 * It will repaint the map after every  mouse drag event, using a frozen image,
 * not refreshing the map truly.
 * 
 * When the mouse is lifted, the extent will be updated.
 * 
 * @see es.prodevelop.gvsig.mobile.gui.tool.behavior.RectangleBehavior
 * @see es.prodevelop.gvsig.mobile.gui.tool.util.PanRefreshTimerTask
 * 
 * @author jldominguez
 *
 */
public class PanListener implements PointToolListener {
	
	private static Logger logger = Logger.getLogger(PanListener.class);
	private static final int REFRESH_MILISECONDS = 100;

	private MapControl mapCtrl;
	private Point startPoint = null;
	private Timer refreshTimer;
	private PanRefreshTimerTask refreshTask;
	private Image frozenImage = null;
	
	/**
	 * Constructor.
	 * 
	 * @param _mc the <code>MapControl</code> needed to perform repaint requests
	 * and reset the extent.
	 */
	public PanListener(MapControl _mc) {
		mapCtrl = _mc;
		
	}

	/**
	 * This method will be called on every point event. If it's a drag event,
	 * the map will be repainted with a slight accumulative offset.
	 * 
	 * @param pevent the point event
	 */
	public void point(PointEvent pevent) {
		
		int id = pevent.getEvent().getID();
		int dif_x = 0;
		int dif_y = 0;

		switch (id) {
		
		case MouseEvent.MOUSE_DRAGGED:

			if (startPoint != null) {
				dif_x = pevent.getPoint().x - startPoint.x;
				dif_y = pevent.getPoint().y - startPoint.y;
				
				logger.debug("MOUSE_DRAGGED - dif_x = " + dif_x + ", dif_y = " + dif_y);

				refreshTask.setOffSetX(dif_x);
				refreshTask.setOffSetY(dif_y);
			}
			break;
			
		case MouseEvent.MOUSE_PRESSED:
			startPoint = pevent.getPoint();
			// frozenImage = mapCtrl.getCurrentImage();
			// int w = frozenImage.getWidth(null);
			// int h = frozenImage.getHeight(null);

			mapCtrl.setToolsImageForPanDrag(true);
			refreshTask = new PanRefreshTimerTask(mapCtrl);
			refreshTask.setOffSetX(0);
			refreshTask.setOffSetY(0);
			refreshTimer = new Timer();
			refreshTimer.schedule(refreshTask, REFRESH_MILISECONDS, REFRESH_MILISECONDS);
			break;
		
		case MouseEvent.MOUSE_RELEASED:

			refreshTimer.cancel();
			
			mapCtrl.setToolsImageForPanDrag(false);

			if (startPoint == null) break;
			
			Rectangle2D oldext = mapCtrl.getViewPort().getAdjustedExtent();
			if (oldext == null) {
				break;
			}
			
			// -------------------------- starts working
			if (mapCtrl.isProcessing()) {
				return;
			} else {
				mapCtrl.setProcessing(true);
			}
			
			mapCtrl.setOffsetAndRefresh(0, 0, false);
			
			dif_x = pevent.getPoint().x - startPoint.x;
			dif_y = pevent.getPoint().y - startPoint.y;
			startPoint = null;
			
			double scalex = mapCtrl.getViewPort().getAffineTransform().getScaleX();
			double scaley = mapCtrl.getViewPort().getAffineTransform().getScaleY();
			double difxgeo = dif_x / scalex;
			double difygeo = dif_y / scaley;

			
			Rectangle2D.Double newext = new Rectangle2D.Double(
					oldext.getMinX() - difxgeo,
					oldext.getMinY() - difygeo,
					oldext.getWidth(),
					oldext.getHeight());

			mapCtrl.getViewPort().setExtent(newext, true);
			break;
		}
	}

	/**
	 * Gets the cursor that should be used while this tool is active.
	 * 
	 * @return the cursor that should be used while this tool is active.
	 */
	public Cursor getCursor() {
		return Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR);
	}

	/**
	 * This method asks the listener to cancel the current drawing process
	 * (currently unsued)
	 * 
	 * @return whether it was possible to cancel the drawing
	 */
	public boolean cancelDrawing() {
		return false;
	}

	/**
	 * Sets the <code>MapControl</code> object that will be updated by
	 * this listener.
	 */
	public void setMapControl(MapControl _mc) {
		mapCtrl = _mc;
	}

}
