/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.driver.raster.ecw;

import es.prodevelop.gvsig.mobile.fmap.util.Utils;

/**
 * Utility class to access Ermapper's libraries through JNI
 * 
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.ecw.EcwRasterDriver
 * 
 * @author jldominguez
 *
 */
public class EcwReader {
	
	/**
	 * Opens ECW file
	 * 
	 * @param full_file_path ecw full file path
	 * @return the file handler as an long value, or zero if the file was not opened
	 * 
	 * @throws Throwable
	 */
	public static native long openEcwFile(String full_file_path) throws Throwable;
	
	/**
	 * Closes the ECW file.
	 * 
	 * @param ptr handler of the previously opened file
	 * @return whether the file was successfully closed
	 */
	public static native boolean closeEcwFile(long ptr);
	
	/**
	 * Renders the ECW raster as an array of RGB values (one int value for each pixel) 
	 * 
	 * @param ptr ECW file handler 
	 * @param geo_min_x minimum x in map units of the zone of interest
	 * @param geo_min_y minimum y in map units of the zone of interest
	 * @param geo_w width of the zone of interest in map units
	 * @param geo_h height of the zone of interest in map units
	 * @param filemingeox minimum x in map units of the raster
	 * @param filemingeoy minimum y in map units of the raster
	 * @param filexcell width of each pixels in map units
	 * @param ffileycell  height of each pixels in map units
	 * @param filew width of the raster in pixels
	 * @param fileh height of the raster in pixels
	 * @param out_w width of the output image in pixels
	 * @param out_h height of the output image in pixels
	 * @param band1 first band requested
	 * @param band2 second band requested
	 * @param band3 third band requested
	 * @return array of integers representing the zone of interest
	 */
	public static native int[] getEcwView(
			long ptr,
			double geo_min_x,
			double geo_min_y,
			double geo_w,
			double geo_h,
			
			double filemingeox,
			double filemingeoy,
			double filexcell,
			double ffileycell,
			int filew, int fileh,
			
			int out_w,
			int out_h,
			int band1,
			int band2,
			int band3);
	
	/**
	 * Gets metadata from the ECW raster
	 * @param ptr the file handler
	 * @return array of doubles with the raster metadata
	 */
	public static native double[] getEcwImageInfo(long ptr);
	
	/**
	 * Utility method to convert an array of integers into an
	 * array of bytes in the PNG format.
	 * 
	 * @param bitmap raster as an array of integers
	 * @param w image width in pixels
	 * @param h image height in pixels
	 * @param offset offset in the array of integers
	 * @param has_transp whether the integers include an alpha band
	 * @return  an array of bytes in the PNG format.
	 */
	public static native byte[] paintArray(
			int[] bitmap, int w, int h, int offset, boolean has_transp);
	
	/**
	 * Test method.
	 * 
	 * @return an irrelevant value
	 */
	public static native int test();
	
	/**
	 * Sets the cache that Ermapper's DLL is allowed to use. It must be set
	 * to zero to avoid memory leaks.
	 * 
	 * @param size cache limits in kilobytes 
	 */
	public static native void setCacheInKBytes(int size);

	/**
	 * gets the array of bytes (PNG format) representing a white image of the given size. 
	 * @param w image width
	 * @param h image height
	 * @return an array of bytes (PNG format) representing a white image
	 */
	public static byte[] getWhiteImage(int w, int h) {
		int size = w * h;
		int[] wimage = new int[size];
		for (int i=0; i<size; i++) { wimage[i] = 16777216; }
		return paintArray(wimage, w, h, 0, false);
	}
	
	static {
		
		try {
			System.loadLibrary("NCSUtil");
			System.loadLibrary("NCScnet");
			System.loadLibrary("NCSEcw");
			
			if (Utils.USING_PDA) {
				System.loadLibrary("j9ecw");
				setCacheInKBytes(0);
			} else {
				System.loadLibrary("j2seecw");
			}
			
		} catch (Throwable th) {
			System.err.println("Unable to load Ermapper libraries!");
		}
		
	}
	
	/**
	 * Utility method to force the load of the associated DLL.
	 *
	 */
	public static void touch() { }
	
}
