package es.prodevelop.gvsig.mobile.fmap.driver.raster.wms;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import es.prodevelop.geodetic.utils.conversion.ConversionCoords;
import es.prodevelop.gvsig.mobile.fmap.driver.FMapDriverException;
import es.prodevelop.gvsig.mobile.fmap.util.ResourceReader;
import es.prodevelop.gvsig.mobile.fmap.util.Utils;
import es.prodevelop.gvsig.mobile.fmap.util.cancel.Cancellable;
import es.prodevelop.gvsig.mobile.fmap.util.download.DownloadException;
import es.prodevelop.gvsig.mobile.fmap.util.download.DownloadTask;
import es.prodevelop.gvsig.mobile.fmap.util.download.Downloader;
import es.prodevelop.gvsig.mobile.fmap.viewport.ViewPort;

public class MercatorCatastroWmsDriver extends WmsRasterDriver {
	
	private static Logger logger = Logger.getLogger(MercatorCatastroWmsDriver.class);	
	
	public MercatorCatastroWmsDriver() {
		
		super(
				"http://ovc.catastro.meh.es/Cartografia/WMS/ServidorWMS.aspx",
				"Catastro",
				"Catastro",
				"Catastro",
				"image/png",
				"EPSG:4326",
				new Rectangle2D.Double(
						-18.409876,
						26.275447,
						23.635856,
						18.579913),
				getInfoFormats(),
				"1.1.1",
				"&"); 
	}
	
	private static ArrayList getInfoFormats() {
		ArrayList resp = new ArrayList();
		resp.add("text/html");
		return resp;
	}
	
	public ArrayList getInfo(
			Point p,
			double tolerance,
			Cancellable cancel,
			String srs) throws FMapDriverException {
		
		return super.getInfo(p, tolerance, cancel, srs);
	}
	
	public void draw(Graphics2D g, ViewPort viewPort)
			throws FMapDriverException {

		Rectangle2D geoView = mercatorToGeo(viewPort.getAdjustedExtent());
		int geoImageHeight =
			Math.round(
			(float) Math.ceil( 
			(1.0 * viewPort.getImageHeight() * 
			geoView.getHeight() / viewPort.getAdjustedExtent().getHeight()))); 

		if (!Utils.rectanglesItersect(geoView, getFullExtent())) {
			return;
		}

		URL request = null;

		try {
			request = new URL(buildMapRequest(geoView, viewPort.getImageWidth(), geoImageHeight));
			logger.debug("WMS Request = " + request.toString());
		} catch (MalformedURLException e) {
			logger.error("Bad map request: " + e.getMessage());
			return;
		}

		// File f = null;
		File image_f = null;
		try {
			// f = Downloader.get(request, "" + imageFormat);
			image_f = (File) Downloader.downloadObjectCancel(request,
					leaveExt("" + imageFormat), DownloadTask.FILE, "wms", null, true);
		} catch (DownloadException e) {
			logger.error("While downloading: " + e.getMessage());
			g.setFont(ERROR_IMAGE_FONT);
			g.setColor(Color.RED);
			int y = viewPort.getImageHeight() - 5;
			g.drawString("WMS " + e.getMessage(), 5, y);
			return;
		}

		logger.debug("WMS - received image, byte size = " + image_f.length());
		Image geoImage = ResourceReader.getImage(image_f); 
		image = geoImageToMerImage(
				geoImage,
				viewPort.getAdjustedExtent(),
				geoView,
				viewPort.getImageWidth(),
				viewPort.getImageHeight());

		logger.debug("WMS - received image, image = " + image);
		if (image != null) {
			logger.debug("WMS - received image, image NOT NULL");
			logger.debug("WMS - received image, image width = "
					+ image.getWidth(null));
		} else {
			logger.debug("WMS - received image, image IS  NULL");
		}

		g.drawImage(image, 0, 0, null);
		image_f = null;
	}	
	

	private Image geoImageToMerImage(
			Image _geo_img,
			Rectangle2D merExt,
			Rectangle2D geoExt,
			int mer_img_w,
			int mer_img_h) {
		
		int geo_img_h = _geo_img.getHeight(null);
		BufferedImage geo_img = Utils.createBufferedImage(
				_geo_img.getWidth(null),
				geo_img_h, BufferedImage.TYPE_INT_ARGB);
		geo_img.getGraphics().drawImage(_geo_img, 0, 0, null);
		
		BufferedImage merImage = Utils.createBufferedImage(
				mer_img_w,
				mer_img_h,
				BufferedImage.TYPE_INT_ARGB);
		
		double geo_pix_h = geoExt.getHeight() / geo_img_h;
		double mer_pix_h = merExt.getHeight() / mer_img_h;
		double top_mer = 0, top_geo = 0;
		double mer_y = 0;
		double[] geo_item;
		int geo_y = 0;
		
		for (int top = 0; top < mer_img_h; top++) {
			top_mer = top * mer_pix_h;
			mer_y = merExt.getMaxY() - top_mer;
			geo_item = ConversionCoords.newmercator2geo(0, mer_y);
			top_geo = geoExt.getMaxY() - geo_item[1];
			geo_y = Math.round((float) Math.floor(top_geo / geo_pix_h));
			
			for (int j=0; j<mer_img_w; j++) {
				int geo_pixel = geo_img.getRGB(j, geo_y);
				if (!isTransp(geo_pixel)) {
					merImage.setRGB(j, top, geo_pixel);
				}
				
			}
		}
		
		return merImage;
	}

	private boolean isTransp(int p) {
		int aux = ((p >> 24) & 0xff);
		return (aux == 0);
	}

	private Rectangle2D mercatorToGeo(Rectangle2D ext) {

		double[] geomaxy = ConversionCoords.newmercator2geo(0, ext.getMaxY()); 
		double[] geominy = ConversionCoords.newmercator2geo(0, ext.getMinY());
		return new Rectangle2D.Double(
				ext.getMinX(),
				geominy[1],
				ext.getMaxX()- ext.getMinX(),
				geomaxy[1] - geominy[1]);
	}

	private String buildMapRequest(Rectangle2D ext, int w, int h) {
    	
		String req = "";
		req = req + serverUrl + "?REQUEST=GetMap&SERVICE=WMS&VERSION=1.1.1&";
		req = req + getPartialQuery(ext, w, h);
		String resp = Utils.replaceAllString(req, " ", "%20");
		logger.debug("buildMapRequest = " + resp);
		return resp;
    }
	
    private String getPartialQuery(Rectangle2D ext, int w, int h) {

        String req = "";
        
        req = "LAYERS=" + requestedLayers + 
           "&SRS=" + srs + 
           "&BBOX=" + ext.getMinX()+ "," +  
           ext.getMinY()+ "," + 
           ext.getMaxX()+ "," + 
           ext.getMaxY() + 
           "&WIDTH=" + w + 
           "&HEIGHT=" + h + 
           "&FORMAT=" + imageFormat.toLowerCase() + 
           "&STYLES=&TRANSPARENT=TRUE";

        return req;
    }	

}
