/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.driver.raster.wms;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

import es.prodevelop.gvsig.mobile.fmap.driver.FMapDriverException;
import es.prodevelop.gvsig.mobile.fmap.driver.raster.GeorreferencedRasterDriver;
import es.prodevelop.gvsig.mobile.fmap.driver.raster.RasterDriver;
import es.prodevelop.gvsig.mobile.fmap.layer.InfoByPoint;
import es.prodevelop.gvsig.mobile.fmap.proj.IProjection;
import es.prodevelop.gvsig.mobile.fmap.util.ResourceReader;
import es.prodevelop.gvsig.mobile.fmap.util.Utils;
import es.prodevelop.gvsig.mobile.fmap.util.cancel.Cancellable;
import es.prodevelop.gvsig.mobile.fmap.util.download.DownloadException;
import es.prodevelop.gvsig.mobile.fmap.util.download.DownloadTask;
import es.prodevelop.gvsig.mobile.fmap.util.download.Downloader;
import es.prodevelop.gvsig.mobile.fmap.util.string.StringUtilities;
import es.prodevelop.gvsig.mobile.fmap.viewport.ViewPort;

/**
 * This is the WMS driver. It is instantiated after the suer has used the WMS dialog.
 * 
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WMSLayerDescription
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WMSProtocolHandler
 * 
 * @author jldominguez
 *
 */
public class WmsRasterDriver implements GeorreferencedRasterDriver, InfoByPoint {
	
	private static Logger logger = Logger.getLogger(WmsRasterDriver.class);	
	public static final Font ERROR_IMAGE_FONT = new Font("Dialog", 1, 12);	
	
	protected Image image = null;
	protected URL serverUrl = null;
	protected String requestedLayers = ""; 
	protected String requestedInfoLayers = null; 
	protected String requestedInfoLayerNames = null; 
	protected String imageFormat = ""; 
	protected String srs = "";
	
	protected String infoFormat = "text/xml";
	protected boolean infoable = false;
	
	protected Rectangle2D fext = null; 
	
	protected IProjection proj = null;

	protected ViewPort lastViewPort = null;
	protected String version = WMSProtocolHandler.DEFAULT_WMS_VERSION;
	protected String symbol = "?";
	
	/**
	 * Constructor. All these paremeters come from the WMS dialog.
	 * 
	 * @param svrUrl the server URL
	 * @param lyrs a comma separated string with the layer ids
	 * @param info_lyrs a comma separated string with the ids of the infoable layers 
	 * @param info_lyrs_names a comma separated string with the names of the infoable layers
	 * @param imgf the image format to be requested
	 * @param _srs the SRS to be used in the requests
	 * @param fe the full extent of the layer, provided by the capabilities response
	 * @param info_fmts a list of the info formats supported by the server
	 * @param v the version  of the WMS protocol to be used
	 * @param symb the separator symbol to be used in the request URLs
	 */
	public WmsRasterDriver(
			String svrUrl,
			String lyrs,
			String info_lyrs,
			String info_lyrs_names,
			String imgf,
			String _srs,
			Rectangle2D fe,
			ArrayList info_fmts,
			String v,
			String symb) {
		
		try {
			serverUrl = new URL(svrUrl);
		} catch (MalformedURLException e) {
			logger.error("Bad server URL: " + svrUrl);
		}
		requestedLayers = lyrs;
		requestedInfoLayers = info_lyrs;
		requestedInfoLayerNames = info_lyrs_names;
		imageFormat = imgf;
		srs = _srs;
		fext = fe;
		
		// decide info format
		if ((info_fmts != null) && (info_fmts.size() > 0)) {
			infoable = true;
			infoFormat = (String) info_fmts.get(0);
			for (int i=0; i<info_fmts.size(); i++) {
				if (((String) info_fmts.get(i)).compareToIgnoreCase("text/xml") == 0) {
					infoFormat = "text/xml";
				}
			}
		}
		
		version  = v;
		symbol = symb;
		
	}
	
	/**
	 * 
	 * @return the WMS protocol version to be used
	 */
	public String getVersion() {
		return version;
	}

	/**
	 * 
	 * @return the separator symbol to be used in the request URLs
	 */
	public String getSymbol() {
		return symbol;
	}

	/**
	 * Initializes the driver with the given projection
	 * @param p the projection used by this driver
	 */
	public void initialize(IProjection p) throws IOException {
		proj = p;
	}

	/**
	 * @return the projection used by this driver
	 */
	public IProjection getProjection() {
		return proj;
	}

	/**
	 * Sets a new projection
	 * @param p the new projection
	 */
	public void setProjection(IProjection p) {
		proj = p;
	}

	/**
	 * @return the full extent of this layer
	 */
	public Rectangle2D getFullExtent() {
		return fext;
	}

	/**
	 * @return whether this is a georreferenced image (return strue)
	 */
	public boolean isGeoreferenced() {
		return true;
	}

	/**
	 * Sets the affine transform associated with this driver. Currently unused.
	 */
	public void setAffineTransform(AffineTransform t) {
	}

	/**
	 * @return the affine transform associated with the last viewport used
	 */
	public AffineTransform getAffineTransform() {
		return lastViewPort.getAffineTransform();
	}

	/**
	 * @return whether the file is accepted as a raster file. Returns false, because
	 * it makes no sense in a WMS driver.
	 */
	public boolean fileAccepted(File fName) {
		return false;
	}

	/**
	 * Not used.
	 */
	public void open(File f) throws IOException {
	}

	/**
	 * Initializes the driver. Currently unused.
	 */
	public void initialize() throws IOException {
		// TODO Auto-generated method stub

	}

	/**
	 * Closes the driver. Tries to prevent memory leaks. 
	 */
	public void close() throws IOException {
		image = null;
		System.gc();
	}

	/**
	 * @return the raster file file path. Returns null, has no sense here.
	 */
	public String getFilePath() {
		return null;
	}

	/**
	 * @return the raster image band count
	 */
	public int getNumBands() {
		return 3;
	}

	/**
	 * @return the transparency level of the image associated with this driver
	 */
	public int getTransparency() {
		return 0;
	}

	/**
	 * Sets the transparecncy level. Currently unused.
	 */
	public void setTransparency(int trans) {
	}

	/**
	 * Draws the image on the given Graphics object.
	 * 
	 * @param g the graphics object on which the image must be drawn
	 * @param viewPort the current viewport object
	 */
	public void draw(Graphics2D g, ViewPort viewPort)
			throws FMapDriverException {
		
		lastViewPort = viewPort;

		if (!Utils.rectanglesItersect(viewPort.getAdjustedExtent(), getFullExtent())) {
			return;
		}
		
		URL request = null;;
			
		try {
			request = new URL(buildMapRequest(viewPort));
			logger.debug("WMS Request = " + request.toString());
		} catch (MalformedURLException e) {
			logger.error("Bad map request: " + e.getMessage());
			return;
		}
		
		// File f = null;
		File image_f = null;
		try {
			// f = Downloader.get(request, "" + imageFormat);
			
				image_f = (File) Downloader.downloadObjectCancel(
					request,
					leaveExt("" + imageFormat),
					DownloadTask.FILE,
					"wms",
					null, true);

		} catch (DownloadException e) {
			logger.error("While downloading: " + e.getMessage());
			g.setFont(ERROR_IMAGE_FONT);
			g.setColor(Color.RED);
			int y = viewPort.getImageHeight() - 5;
			g.drawString("WMS " + e.getMessage(), 5, y);
			return;
		}
		
		logger.debug("WMS - received image, byte size = " + image_f.length());
		image = ResourceReader.getImage(image_f);
		// image = ResourceReader.getResourceImage("test", "test.png");
		// .getImage(new File("\\test.jpg"));
		
		logger.debug("WMS - received image, imagee = " + image);
		if ((image != null) && (image.getWidth(null) != -1)) {
			logger.debug("WMS - received image, image width = " + image.getWidth(null));
			g.drawImage(image, 0, 0, null);
		} else {
			logger.debug("WMS - received image, image IS  NULL");
			g.setFont(ERROR_IMAGE_FONT);
			g.setColor(Color.RED);
			int y = viewPort.getImageHeight() - 5;
			g.drawString("WMS image error", 5, y);
		}
		image_f = null;
	}

	protected String leaveExt(String str) {
		
		int ind_bar = str.indexOf("/"); 
		if (ind_bar == -1) {
			return str;
		} else {
			return str.substring(ind_bar + 1, str.length());
		}
	}

	/**
	 * @return the data type of the individual bands 
	 */
	public int getRasterDataType() {
		// TODO Auto-generated method stub
		return 0;
	}

	/**
	 * Gets the data associated with the given pixel and band
	 * 
	 * @param x the x coordinate of the pixel of interest
	 * @param y the y coordinate of the pixel of interest
	 * @param band
	 */
	public Object getData(int x, int y, int band) {
		// TODO Auto-generated method stub
		return null;
	}

	/**
	 * Gets the RGB values of the pixel for the given coordinates (in map units)
	 * 
	 * @param wcx x coordinate of the point of interest (map units)
	 * @param wcy y coordinate of the point of interest (map units)
	 * 
	 * @return the RGB values of the pixel
	 * 
	 */
	public int[] getPixel(double wcx, double wcy) {
		
		int[] resp = null;
		
		if (image == null) {
			
			resp = new int[1];
			resp[0] = RasterDriver.ONE_PIXEL_NO_DATA;
			
		} else {
			
			logger.debug("WMS DRIVER - getPixel, wx = " + wcx);
			logger.debug("WMS DRIVER - getPixel, wy = " + wcy);
			
			Point2D wp = new Point2D.Double(wcx, wcy);
			Point2D screenp = new Point2D.Double();
			lastViewPort.getAffineTransform().transform(wp, screenp);
			int scx = (int) Math.round(screenp.getX());
			int scy = (int) Math.round(screenp.getY());
			
			logger.debug("WMS DRIVER - getPixel, scx = " + scx);
			logger.debug("WMS DRIVER - getPixel, scy = " + scy);

			if ((scx >= image.getWidth(null)) || (scy >= image.getHeight(null))) {
				
				resp = new int[1];
				resp[0] = RasterDriver.ONE_PIXEL_NO_DATA;
				
			} else {

				BufferedImage bi = Utils.createBufferedImage(1, 1);
				bi.getGraphics().drawImage(image, -scx, -scy, null);
				resp = new int[3];
				
				int pixel = bi.getRGB(0, 0); 
				logger.debug("WMS DRIVER - pixel = " + pixel);
				resp[0] = (pixel >> 16) & 0xff;
				resp[1] = (pixel >> 8) & 0xff;
				resp[2] = pixel & 0xff;

			}
		}
		return resp;
	}

	/**
	 * @return the driver name
	 */
	public String getName() {
		return "WMS Raster Driver";
	}
	
    private String buildMapRequest(ViewPort vp) {
    	
		String req = serverUrl.toString() + symbol + "REQUEST=GetMap&SERVICE=WMS&VERSION=" + version + "&";
		req = req + getPartialQuery(vp);
		String resp = Utils.replaceAllString(req, " ", "%20");
		logger.debug("buildMapRequest = " + resp);
		return resp;
    }
    
    private String buildFeatureInfoRequest(ViewPort vp, int screen_x, int screen_y, String lyr_id) {
    	
    	String req = "";
    	Rectangle2D ext = vp.getAdjustedExtent();
    	
		req = req + serverUrl + symbol + "REQUEST=GetFeatureInfo&SERVICE=WMS&VERSION=" + version;
		
		req = req + "&QUERY_LAYERS=" + lyr_id;
		req = req + "&LAYERS=" + lyr_id;
		req = req + "&INFO_FORMAT=" + infoFormat;
		req = req + "&SRS=" + srs;
		req = req + "&BBOX=" + ext.getMinX()+ "," + ext.getMinY()+ "," + ext.getMaxX()+ "," + ext.getMaxY();
		
		req = req + "&WIDTH=" + vp.getImageWidth();
		req = req + "&HEIGHT=" + vp.getImageHeight();
		
		req = req + "&FORMAT=" + imageFormat.toLowerCase();
		req = req + "&STYLES=" + "";
		req = req + "&TRANSPARENT=" + "";
		req = req + "&X=" + screen_x;
		req = req + "&Y=" + screen_y;
		req = req + "&FEATURE_COUNT=" + "1";
		req = req + "&EXCEPTIONS=" + "XML";
		
		return req;
    }
    
    private String getPartialQuery(ViewPort vp) {

        String req = "";
        
        Rectangle2D ext = vp.getAdjustedExtent();
        
        req = "LAYERS=" + requestedLayers + 
           "&SRS=" + srs + 
           "&BBOX=" + ext.getMinX()+ "," +  
           ext.getMinY()+ "," + 
           ext.getMaxX()+ "," + 
           ext.getMaxY() + 
           "&WIDTH=" + vp.getImageWidth() + 
           "&HEIGHT=" + vp.getImageHeight() + 
           "&FORMAT=" + imageFormat.toLowerCase() + 
           "&STYLES=&TRANSPARENT=TRUE";

        return req;
    }
    
    private void appendInfoForLayer(Point p, String lyr_id, String lyr_name, ArrayList more_info, boolean write_lyr_name) {
    	
		URL request = null;
		
		try {
			request = new URL(buildFeatureInfoRequest(lastViewPort, p.x, p.y, lyr_id));
			logger.debug("WMS GetFeatureInfo, request = " + request.toString());
		} catch (MalformedURLException e) {
			logger.error("Bad map request: " + e.getMessage());
			more_info.add("[Bad request]");
			more_info.add(null);
		}
		
		more_info.add(null); more_info.add(null);
		if (write_lyr_name) {
			more_info.add(lyr_name);
			more_info.add(null);
		}

		File feature_info_f = null;
		try {
			feature_info_f = (File) Downloader.downloadObjectCancel(
					request,
					"inf",
					DownloadTask.FILE,
					"wms",
					null, true);
		} catch (DownloadException e) {
			logger.error("While downloading: " + e.getMessage());
			more_info.add("[" + e.getMessage() + "]");
			more_info.add(null);
		}
		more_info.addAll(parseFeatureInfoResponse(feature_info_f));
    }

    
    /**
     * get info associated with the given screen pixel.
     * 
     * @param p the screen point of interest
     * @param tolerance tolerance ised for query width (unused here)
     * @param cancel cancel object
     * @param srs used SRS
     *
     * @return
     */
	public ArrayList getInfo(
			Point p,
			double tolerance,
			Cancellable cancel,
			String srs) throws FMapDriverException {
		
		ArrayList resp = new ArrayList();
		
		if ((infoable) && (requestedInfoLayers != null)) {
			
			String[] query_lyr_arr = StringUtilities.split(requestedInfoLayers, ",");
			String[] query_lyr_names_arr = StringUtilities.split(requestedInfoLayerNames, ",");
			
			if (query_lyr_arr.length > 10) {
				
				// sorry, too many layers, it has to be all in one :(
				appendInfoForLayer(p, requestedInfoLayers, null,  resp, false);
				
			} else {
				
				for (int i=0; i<query_lyr_arr.length; i++) {
					appendInfoForLayer(p, query_lyr_arr[i], query_lyr_names_arr[i], resp, true);
				}
				
			}

		} else {
			
			resp.add("[Layer not queryable]");
			resp.add(null);

		}

		return resp;
	}
	
	
    private ArrayList parseFeatureInfoResponse(File f) {
    	
    	if ((!f.exists()) || (f.length() < 5)) {
    	
    		ArrayList empty = new ArrayList();
    		logger.error("Feature info response is empty (possibly due to valid but too long request)");
    		empty.add("[Empty response]");
    		empty.add(null);
    		return empty;
    	}
    	
    	boolean htmlCode = false;
    	boolean xmlCode = true;

    	ArrayList resp = new ArrayList();
    	String enc = "ISO-8859-1";

    	FileReader reader = null;
    	try {
    		reader = new FileReader(f);
    		char[] buffer = new char[200];
    		reader.read(buffer);
    		StringBuffer strBuff = new StringBuffer();
    		strBuff.append(buffer);
    		String auxStr = strBuff.toString().toLowerCase();
	    	htmlCode = (auxStr.indexOf("<html") != -1);
	    	xmlCode = (auxStr.indexOf("<?xml") != -1);
            int a = auxStr.indexOf("encoding=");
            if (a != -1) {
            	auxStr = auxStr.substring(a, auxStr.length());
            	a = auxStr.indexOf("\"");
            	auxStr = auxStr.substring(a + 1, auxStr.length());
            	int b = auxStr.indexOf("\"");
            	enc = auxStr.substring(0, b);
            }
            
            reader.close();
            
    	} catch (Exception ex)	{
    		logger.error("In main parse method: " + ex.getMessage());
    		resp.add("[Parse error]");
    		resp.add(null);
    		return resp;
		}
    	
    	if ((!htmlCode) && (!xmlCode)) {
    		// plain text
    		return parseAsPlainText(f);
    	}

    	int tag;
    	String lastEntity = "";
    	KXmlParser kxmlParser = null;
    	kxmlParser = new KXmlParser();
    	
    	try {
    		
    		kxmlParser.setInput(new FileInputStream(f), enc);
    		if (htmlCode) setHtmlReplacements(kxmlParser);

			tag = kxmlParser.nextTag();
			
    		if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {

				 while (tag != KXmlParser.END_DOCUMENT) {
                     switch (tag) {

						case KXmlParser.START_TAG:
							
							lastEntity = kxmlParser.getName();
							if (isRelevantEntity(lastEntity)) {
								for (int i=0; i<kxmlParser.getAttributeCount(); i++) {
									String attname = kxmlParser.getAttributeName(i); 
									if (isRelevantEntity(attname)) {
										addPairRemovingTitles(
												resp,
												attname,
												kxmlParser.getAttributeValue(i));
									}
								}
							} else {
								logger.warn("Found non relevant ENTITY: " + lastEntity);
							}

							break;
						case KXmlParser.END_TAG:
							break;
						case KXmlParser.TEXT:
							
							String txt = kxmlParser.getText();

							if (isRelevantEntity(lastEntity) && isRelevantString(txt)) {
								addPairRemovingTitles(resp, lastEntity, txt);
							} else {
								logger.warn("Found non relevant ENTITY/TEXT: " + lastEntity + ", " + txt);
							}
							break;
					 }
                     
                     try {
        				 tag = kxmlParser.next();
                     } catch (XmlPullParserException pex) {
                    	 logger.debug("ERROR: " + pex.getMessage());
                     }
                     
    			 }
    			kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
    		}
    	} catch (XmlPullParserException parser_ex){
    		logger.error("While parsing: " + parser_ex.getMessage());
    		addPairRemovingTitles(resp, "[Parse error]", null);
    		return resp;
    	} catch (IOException ioe) {
    		logger.error("While parsing: " + ioe.getMessage());
    		addPairRemovingTitles(resp, "[I/O error]", null);
    		return resp;
        }
    	
		return resp;
    }
    
    private ArrayList parseAsPlainText(File f) {

    	ArrayList resp = new ArrayList();
    	
    	// FileReader reader = null;
    	try {
    		FileInputStream fis = new FileInputStream(f);
    		InputStreamReader isr = new InputStreamReader(fis, "ISO-8859-1");

    		char[] buffer = new char[20];
    		StringBuffer strBuff = new StringBuffer();
    		
    		while (isr.read(buffer) != -1) {
    			strBuff.append(buffer);
    		}
    		
    		isr.close();
    		fis.close();

    		String[] lines = Utils.splitString(strBuff.toString(), "\r\n");
    		for (int i=0; i<lines.length; i++) {
    			breakLineAndAdd(resp, lines[i]);
    		}
    	} catch (Exception ex)	{
    		logger.error("While parsing plain text: " + ex.getMessage());
    		resp.clear();
    		resp.add("[Parse error]");
    		resp.add(null);
		}    	
		return resp;
	}

	private void breakLineAndAdd(ArrayList more_info, String str) {
		
		if (str == null) return;
		int att = str.indexOf("=");
		if (att != -1) {
			
			breakHereAndAdd(more_info, str, att);
			
		} else {
			
			att = str.indexOf(":");
			if (att != -1) {
				breakHereAndAdd(more_info, str, att);
			} else {
				
				att = str.indexOf(" ");
				if (att != -1) {
					breakHereAndAdd(more_info, str, att);
				} else {
					more_info.add("");
					more_info.add(str);
				}
			}
		}
	}

	private void breakHereAndAdd(ArrayList more_info, String str, int here) {
		more_info.add(str.substring(0, here));
		more_info.add(str.substring(here + 1));
	}

	private void addPairRemovingTitles(ArrayList al, String left, String right) {
    	
    	if (left.compareToIgnoreCase("p") == 0) {
    		al.add("");
    	} else {
        	if (left.compareToIgnoreCase("href") == 0) {
        		al.add("");
        	} else {
            	if (left.compareToIgnoreCase("a") == 0) {
            		al.add("Data :");
            	} else {
            		al.add(left + " :");
            	}
        	}
    	}
    	
    	al.add(right);
    }

	private boolean isRelevantEntity(String str) {
		if (str.compareToIgnoreCase("lang") == 0) return false;
		if (str.toLowerCase().indexOf("xml") != -1) return false;
		return true;
	}
	
	private boolean isRelevantString(String str) {
		
		if (str == null) return false;
		
		int l = str.length(); 
		if (l == 0) return false;
		if (l > 8) return true;

		byte[] str_bytes = str.getBytes();
		int bytes_l = str_bytes.length;

		// 32-space, 10-lf, 13-cr
		if (str_bytes[0]           == 32) return isRelevantString(str.substring(1, l));
		if (str_bytes[bytes_l - 1] == 32) return isRelevantString(str.substring(0, l - 1));
		if (str_bytes[0]           == 10) return isRelevantString(str.substring(1, l));
		if (str_bytes[bytes_l - 1] == 10) return isRelevantString(str.substring(0, l - 1));
		if (str_bytes[0]           == 13) return isRelevantString(str.substring(1, l));
		if (str_bytes[bytes_l - 1] == 13) return isRelevantString(str.substring(0, l - 1));
		
		return true;
	}
	
	private void setHtmlReplacements(KXmlParser p) {
		
		logger.debug("Setting replacements...");

		try {

			// -----------
			// Latin 1 entities
			// -----------

			p.defineEntityReplacementText("nbsp", "\u00a0");
			p.defineEntityReplacementText("iexcl", "\u00a1");
			p.defineEntityReplacementText("cent", "\u00a2");
			p.defineEntityReplacementText("laquo", "\u00ab");
			p.defineEntityReplacementText("acute", "\u00b4");
			p.defineEntityReplacementText("iquest", "\u00bf");

			p.defineEntityReplacementText("Agrave", "\u00c0");
			p.defineEntityReplacementText("Aacute", "\u00c1");
			p.defineEntityReplacementText("Ccedil", "\u00c7");
			p.defineEntityReplacementText("Egrave", "\u00c8");
			p.defineEntityReplacementText("Eacute", "\u00c9");
			p.defineEntityReplacementText("Igrave", "\u00cc");
			p.defineEntityReplacementText("Iacute", "\u00cd");
			p.defineEntityReplacementText("Ntilde", "\u00d1");
			p.defineEntityReplacementText("Ograve", "\u00d2");

			p.defineEntityReplacementText("Oacute", "\u00d3");
			p.defineEntityReplacementText("Ugrave", "\u00d9");

			p.defineEntityReplacementText("Uacute", "\u00da");
			p.defineEntityReplacementText("Yacute", "\u00dd");
			p.defineEntityReplacementText("agrave", "\u00e0");
			p.defineEntityReplacementText("aacute", "\u00e1");

			p.defineEntityReplacementText("ccedil", "\u00e7");
			p.defineEntityReplacementText("egrave", "\u00e8");
			p.defineEntityReplacementText("eacute", "\u00e9");

			p.defineEntityReplacementText("igrave", "\u00ec");
			p.defineEntityReplacementText("iacute", "\u00ed");
			p.defineEntityReplacementText("ntilde", "\u00f1");
			p.defineEntityReplacementText("ograve", "\u00f2");
			p.defineEntityReplacementText("oacute", "\u00f3");
			p.defineEntityReplacementText("ugrave", "\u00f9");
			p.defineEntityReplacementText("uacute", "\u00fa");
			p.defineEntityReplacementText("yacute", "\u00fd");

			// -----------
			// Special entities
			// -----------

			p.defineEntityReplacementText("euro", "\u20ac");

		} catch (XmlPullParserException e) {
			logger.error("While defining replacements: " + e.getMessage());
		}
	}

	/**
	 * 
	 * @return the server URL
	 */
	public URL getServerUrl() {
		return serverUrl;
	}

	/**
	 * 
	 * @return comma separated string of requested layers
	 */
	public String getRequestedLayers() {
		return requestedLayers;
	}
	
	/**
	 * 
	 * @return comma separated string of requested styles, one for each layer
	 */
	public String getRequestedStyles() {
		return "";
	}

	/**
	 * 
	 * @return the requested image format
	 */
	public String getImageFormat() {
		return imageFormat;
	}

	/**
	 * 
	 * @return the requested SRS
	 */
	public String getSrs() {
		return srs;
	}
	


}
