/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.driver.vect.shp;

import es.prodevelop.gvsig.mobile.fmap.util.Utils;

/**
 * This class accesses the SHP/DBF file via static JNI functions.
 * 
 * @author jldominguez
 *
 */
public class ShpReader {
	
	static {
		
		try {
			if (Utils.USING_PDA) {
				System.loadLibrary("j9shp");
			} else {
				System.loadLibrary("j2seshp");
			}
		} catch (Throwable th) {
			System.err.println("Unable to load SHP libraries!");
		}

	}

	/**
	 * Opens a SHP file
	 * 
	 * @param full_file_path file path
	 * @param write_access whether a write access is needed
	 * 
	 * @return the handler to the opened file
	 */
	public static native long openShpFile(String full_file_path, boolean write_access);
	
	/**
	 * Closes a SHP file
	 * @param ptr the handler of the file to be closed
	 * @return whether the file was successfully closed
	 */
	public static native boolean closeShpFile(long ptr);

	/**
	 * Gets SHP metadata
	 * @param ptr handler to the SHP file
	 * @return an array containing metadata: [0] = shape type, [1] = extent min x, 
	 * [2] = extent min y, [3] = extent width, [4] = extent height, [5] = row count,
	 * 
	 */
	public static native double[] getShpMetadata(long ptr);

	/**
	 * Gets the shape as an array of coordinates
	 * @param ptr handler of the SHP file
	 * @param index index of the geometry of interest
	 * @return the shape as an array of coordinates: [minx, miny, w, h, type,
	 * n_parts, n_parts_start_index, nvertices, vertices]
	 */
	public static native double[] getShapeWithBox(long ptr, int index);
	
	/**
	 * Gets the shapebounds of a geometry
	 * @param ptr handler of the SHP file
	 * @param index index of the geometry of interest
	 * @return bounding box [x, y, w, h]
	 */
	public static native double[] getShapeBounds(long ptr, int index);
	
	/**
	 * Utility method to transform a geometry with a given affine transform.
	 * This method is incomplete and should not be used.
	 * 
	 * @param data the corrdinates to be transformed
	 * @param size the number of vertices
	 * @param m00 component of the affine transform matrix
	 * @param m11 component of the affine transform matrix
	 * @param m02 component of the affine transform matrix
	 * @param m12 component of the affine transform matrix
	 * @param dim the dimension of the geometry
	 * @return
	 */
	public static native int[] getPixelTransformedShape(
			double[] data,
			int size,
			double m00,
			double m11,
			double m02,
			double m12,
			int dim);

	/**
	 * Opens a DBF file.
	 *  
	 * @param canonicalPath file path 
	 * @param write_access whether write access is needed
	 * @return the handler to the file
	 */
	public static native long openDbfFile(String canonicalPath, boolean write_access);
	
	/**
	 * Gets the field count of a DBF file
	 * @param fileHandler teh handler of the file of interest
	 * @return the field count
	 */
	public static native int getDbfFileFieldCount(long fileHandler);
	
	/**
	 * Gets the field types of a DBF file
	 * @param fileHandler the handler of the file of interest
	 * @param size field count
	 * @return the field types as an array of chars (treated as integers)
	 */
	public static native char[] getDbfFileFieldTypes(long fileHandler, int size);
	
	/**
	 * Gets the field names of a DBF file
	 * @param fileHandler the handler of the file of interest
	 * @param size  field count
	 * @return the field types as a comma separated list 
	 */
	public static native String getDbfFileFieldNames(long fileHandler, int size);
	
	/**
	 * Closes the file 
	 * @param fileHandler the handler of the DBF file to be closed
	 * @return whether the file was successfully closed
	 */
	public static native boolean closeDbfFile(long fileHandler);
	
	/**
	 * Gets row count
	 * @param fileHandler the handler of the file of interest
	 * @return the row count
	 */
	public static native int getDbfFileRowCount(long fileHandler);
	
	/**
	 * Gets the field widths of a DBF file
	 * @param fileHandler the handler of the file of interest
	 * @param size field count
	 * @return the field types as a comma separated list 
	 */
	public static native int[] getDbfFileFieldWidths(long fileHandler, int size);
	
	/**
	 * Gets a numeric value from a row and field
	 * 
	 * @param handler the handler of the file of interest
	 * @param row the row index
	 * @param fieldindex the field index
	 * @return the numeric value of that cell (row, field) 
	 */
	public static native double getDbfNumberFieldValue(long handler, int row, int fieldindex);
	
	/**
	 * Gets a string value from a row and field
	 * 
	 * @param handler the handler of the file of interest
	 * @param row the row index
	 * @param fieldindex the field index
	 * @return the string value of that cell (row, field) 
	 */
	public static native String getDbfStringFieldValue(long handler, int row, int fieldindex);
	
	/**
	 * Gets a boolean value from a row and field
	 * 
	 * @param handler the handler of the file of interest
	 * @param row the row index
	 * @param fieldindex the field index
	 * @return the boolean value of that cell (row, field) 
	 */
	public static native boolean getDbfBooleanFieldValue(long handler, int row, int fieldindex);
	
	/**
	 * Utility method to force the load of the DLL library
	 *
	 */
	public static void touch() { }

}
