package es.prodevelop.gvsig.mobile.fmap.edition;

import java.io.IOException;

import com.hardcode.driverManager.Driver;
import com.hardcode.gdbms.engine.data.driver.DriverException;

import es.prodevelop.gvsig.mobile.fmap.core.IRow;
import es.prodevelop.gvsig.mobile.fmap.driver.FMapDriverException;
import es.prodevelop.gvsig.mobile.fmap.driver.FieldDescription;
import es.prodevelop.gvsig.mobile.fmap.driver.ITableDefinition;
import es.prodevelop.gvsig.mobile.fmap.edition.commands.CommandRecord;
import es.prodevelop.gvsig.mobile.fmap.layer.SelectableDataSource;
import es.prodevelop.gvsig.mobile.fmap.util.graph2d.FBitSet;


/**
 * DOCUMENT ME!
 *
 * @author $author$
 */
public interface IEditableSource {
	/**
     * Mtodo invocado cuando se comienza la edicin, para poner en marcha las
     * estructuras de datos necesarias para la misma, notificar al servidor en
     * protocolos en los que sea necesario, ...
	 * @throws StartEditionLayerException
	 * @throws StartWriterVisitorException
     *
     * @throws EditionException Si no se logra poner la fuente de datos en
     *         edicin
     */
    void startEdition(int sourceType) throws IOException,FMapDriverException;

    /**
     * Invocado cuando termina la edicin. En funcin de la clase concreta que
     * implemente este mtodo se generar el fichero con los resultados de la
     * edicin, se realizar una transaccin a la base de datos, etc.
     * @throws StopEditionLayerException
     * @throws StopWriterVisitorException
     *
     * @throws EditionException Si no se consiguen llevar a cabo las
     *         modificaciones
     */
   void stopEdition(IWriter writer, int sourceType) throws IOException;

    /**
     * Cancela la edicin sin escribir los cambios
     * @throws CancelEditingLayerException
     *
     * @throws IOException Si se produce un error
     */
    void cancelEdition(int sourceType) throws IOException;

    /**
     * Si el ndice se corresponde a una geometria de las originales de la capa
     * en edicin y no ha sido modificada ni eliminada devuelve la geometria
     * original. Si ha sido modificada debera de buscar en el fichero de
     * expansin y si ha sido eliminada debera devolver null
     *
     * @param index ndice de la geometra.
     *
     * @return Geometra.
     * @throws ReadDriverException
     * @throws ExpansionFileReadException
     *
     * @throws IOException Si se produce un error con el fichero de expansin
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     */
    IRowEdited getRow(int index) throws IOException;

    /**
     * Devuelve el nmero de geometrias que hay actualmente en edicin.
     *
     * @return Nmero de geometras.
     * @throws ReadDriverException
     *
     * @throws DriverIOException Si se produce un error accediendo a la capa
     * @throws DriverException
     */
    int getRowCount() throws IOException;

    /**
     * Aade una geometria al fichero de expansin y guarda la correspondencia
     * en una tabla asociada.
     *
     * @param g geometra a guardar.
     * @throws ValidateRowException
     * @throws ExpansionFileWriteException
     * @throws ReadDriverException
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
     int addRow(IRow row,String descrip, int sourceType) throws IOException;

    /**
     * Deshace la ltima accin realizada. Si no hay ms acciones no realiza
     * ninguna accin
     * @throws EditionCommandException
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
    void undo() throws IOException;

    /**
     * Rehace la ltima accin deshecha. Si no hay ms acciones no hace nada
     * @throws EditionCommandException
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         filas originales
     */
    void redo() throws IOException;

    /**
     * Elimina una geometria. Si es una geometra original de la capa en
     * edicin se marca como eliminada (haya sido modificada o no). Si es una
     * geometra aadida posteriormente se invalida en el fichero de
     * expansin, para que una futura compactacin termine con ella.
     *
     * @param index ndice de la geometra que se quiere eliminar
     * @throws ExpansionFileReadException
     * @throws ReadDriverException
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
    void removeRow(int index,String descrip, int sourceType) throws IOException;

    /**
     * Si se intenta modificar una geometra original de la capa en edicin se
     * aade al fichero de expansin y se registra la posicin en la que se
     * aadi. Si se intenta modificar una geometria que se encuentra en el
     * fichero de expansin (por ser nueva o original pero modificada) se
     * invoca el mtodo modifyGeometry y se actualiza el ndice de la
     * geometria en el fichero.
     *
     * @param index ndice de la geometra que se quiere eliminar
     * @param type TODO
     * @param g Geometra nueva
     * @throws ExpansionFileWriteException
     * @throws ValidateRowException
     * @throws ReadDriverException
     * @throws ExpansionFileReadException
     * @throws IOException Si se produce un error con el fichero de expansin
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     */
    int modifyRow(int index, IRow row,String descrip, int sourceType) throws IOException;



    /**
     * Compacta el almacenamiento de las geometras que estn en edicin. Tras
     * esta operacin, el orden de las geometras seguramente cambiar y toda
     * llamada a getGeometry devolver una geometra distinta de null, ya que
     * las eliminadas son borradas definitivamente. Hay que tener especial
     * cuidado al invocar este mtodo ya que cualquier tipo de asociacin
     * entre geometras y otro tipo de objetos (comandos de edicin, snapping,
     * ...) que use el ndice de la geometra se ver afectado por ste mtodo
     */
    void compact();

    /**
     * Establece la imagen de las geometras seleccionadas con el fin de que en
     * una edicin interactiva se pueda obtener dicha imagen para simular el
     * copiado, rotado, etc
     *
     * @param i imagen
     */
    // void setSelectionImage(Image i);

    /**
     * Obtiene una imagen con las geometras seleccionadas
     *
     * @return imagen
     */
    // Image getSelectionImage();

    /**
     * DOCUMENT ME!
     */
    void startComplexRow();

    void endComplexRow(String description);
    public void undoModifyRow(int geometryIndex,int previousExpansionFileIndex, int sourceType) throws IOException;
    public IRow doRemoveRow(int index, int sourceType) throws IOException;
    public int doModifyRow(int index, IRow feat, int sourceType) throws IOException;
    public int doAddRow(IRow feat, int sourceType) throws IOException;
    public void undoRemoveRow(int index, int sourceType) throws IOException;
    public void undoAddRow(int index, int sourceType) throws IOException;
    public SelectableDataSource getRecordset() throws FMapDriverException;
    public boolean isEditing();
    public FBitSet getSelection() throws FMapDriverException;
    public CommandRecord getCommandRecord();

	public void addEditionListener(IEditionListener listener);

	public void removeEditionListener(IEditionListener listener);

	public ITableDefinition getTableDefinition() throws IOException;

	public void validateRow(IRow row, int sourceType) throws IOException;

	/**
	 *  Use it to add, remove or rename fields. If null, you cannot modifiy the table structure
	 *  (for example, with dxf files, dgn files, etc).
	 *  The changes will be applied when stopEditing() is called.
	 */
	public IFieldManager getFieldManager();

	public void saveEdits(IWriter writer, int sourceType) throws IOException;

	/**
	 * @return
	 */
	public Driver getOriginalDriver();

	/**
	 * Please, use this if you need support for defaultValues.
	 * Don't user getRecordset().getFieldsDescription()!!.
	 * The reason is ResultSetMetadata has no information about
	 * defaultValues in fields.
	 * @return
	 */
	public FieldDescription[] getFieldsDescription();


}
