/* gvSIG. Geographic Information System of the Valencian Government
*
* Copyright (C) 2007-2008 Infrastructures and Transports Department
* of the Valencian Government (CIT)
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
* MA  02110-1301, USA.
* 
*/

package es.prodevelop.gvsig.mobile.gpe.fmap;

import java.io.IOException;

import com.hardcode.driverManager.DriverLoadException;
import com.hardcode.gdbms.engine.data.DataSource;
import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.values.Value;

import es.prodevelop.gvsig.mobile.fmap.core.DefaultFeature;
import es.prodevelop.gvsig.mobile.fmap.core.IFeature;
import es.prodevelop.gvsig.mobile.fmap.core.IGeometry;
import es.prodevelop.gvsig.mobile.fmap.driver.FMapDriverException;
import es.prodevelop.gvsig.mobile.fmap.driver.VectorialFileDriver;
import es.prodevelop.gvsig.mobile.fmap.layer.SelectableDataSource;
import es.prodevelop.gvsig.mobile.fmap.layer.VectorialAdapter;

/**
 * Adapter for a vectorial file to the vectorial interface, keeping 
 * the needed state by a vectorial file layer.						
 * 																	
 * <br>2008 PRODEVELOP SL</br> 
 * @author Carlos Snchez Perin <br>csanchez@prodevelop.es</br>
 */
public class VectorialGpeAdapter extends VectorialAdapter{
	
	/*
	 * GLOBAL VARIABLES
	 */
	private int reference_count = 0; 			//Call to VectorialGpeAdapter counter only one time is opened.
	private boolean driverInitialized = false; 	//If GpeFileDriver is initialized becomes true.
	private SelectableDataSource ds;			//DataSource that we have chase to read the tables
	
	//CONSTANTES
	public static final String XML_FILE_DATASOURCE = "XML File Data Source";
	public static final String GPE_FILE_DATASOURCE = "GML File Data Source";
	public static final String KML_FILE_DATASOURCE = "KML File Data Source";
	
	
	/**
	 * Synchronized to get the features and allow thread safe access to 
	 * features stored in files.
	 * 
	 * @param number of feature to get
	 */
	public synchronized IFeature getFeature(int numReg) throws FMapDriverException {
        IGeometry geom;
        IFeature feat = null;
        try {
            geom = getShape(numReg);
            DataSource rs = getRecordset();
            Value[] regAtt = new Value[rs.getFieldCount()];
            for (int fieldId=0; fieldId < rs.getFieldCount(); fieldId++ )
            {
                regAtt[fieldId] =  rs.getFieldValue(numReg, fieldId);
            }
            feat = new DefaultFeature(geom, regAtt, "" + numReg);
        } catch (DriverException e) {
            throw new FMapDriverException(e.getMessage());
        } catch (DriverLoadException e) {
            throw new FMapDriverException(e.getMessage());
        }
        return feat;
    }
	
	/**
	 * Set the conditional referenceCount variable into another value.
	 * 
	 * @param counter
	 */
	public void setReferenceCount(int c) {
		reference_count = c;
	}
	
	/**
	 * Returns the DataSource by the file name.
	 * 
	 * @see es.prodevelop.gvsig.mobile.fmap.layer.VectorialAdapter#getRecordset()
	 */
	public SelectableDataSource getRecordset() {
		try {
			if (ds == null) {
				VectorialGpeDriver driver = (VectorialGpeDriver) getFileDriver();
				DataSourceGpe data = new DataSourceGpe(driver.getAttributes(),driver.getPosition(),driver.getFields());
				ds= new SelectableDataSource(data);
			}
		} catch (DriverException e) {
			throw new RuntimeException(e);
		}
		return ds;
	}

	/**
	 * Synchronized to get the geometries and allow thread safe access to 
	 * features stored in files.
	 * 
	 * @throws FMapDriverException  
	 * @see es.prodevelop.gvsig.mobile.fmap.layer.ReadableVectorial#getShape(int)
	 */
	public synchronized IGeometry getShape(int index) throws FMapDriverException {
		try {
			return getFileDriver().getShape(index);
		} catch (IOException e) {
			throw new FMapDriverException(e.getMessage());
		}
	}
	
	/**
	 * Return constant with which we know the type of geometry is.
	 * 
	 * @throws FMapDriverException  
	 * @see es.prodevelop.gvsig.mobile.fmap.layer.ReadableVectorial#getShapeType()
	 */
	public int getShapeType() throws FMapDriverException {
		return getFileDriver().getShapeType();
	}
	
	/**	
	 * Increases the counter every time the file is open.
	 * The file only is opened the first time when the counter is zero.
	 * 
	 * @throws DriverException 
	 * @see es.prodevelop.gvsig.mobile.fmap.layer.ReadableVectorial#start()
	 */
	public synchronized void start() throws DriverException {
	try {
	    if (reference_count == 0)
	    {
			if (!driverInitialized) {
				getFileDriver().initialize();
				driverInitialized = true;
			}
	    }
		reference_count++;
		} catch (IOException e) {
			throw new DriverException(e.getMessage());
		}
	}
	
	/**	
	 * Decreases the counter every time there's a close action.
	 * The file only is closed when the counter is zero.
	 * 
	 * @throws DriverException 
	 * @see es.prodevelop.gvsig.mobile.fmap.layer.ReadableVectorial#stop()
	 */
	public synchronized void stop() throws DriverException {
		try {
		    if (reference_count == 0)
		    {
		        getFileDriver().close();
		    }
		    else
		        if (reference_count < 0)
		            throw new RuntimeException("Contador de referencias de driver ="+ reference_count + ". Demasiados stop().");
	    	reference_count--;
		} catch (IOException e) {
			throw new DriverException(e.getMessage());
		}
	}
	
	/**
	 * Returns the right driver to parse the file.
	 * 
	 * @return VectorialFileDriver
	 */
	private VectorialFileDriver getFileDriver() {
		return (VectorialGpeDriver) getDriver();
	}
}
