/* gvSIG. Geographic Information System of the Valencian Government
*
* Copyright (C) 2007-2008 Infrastructures and Transports Department
* of the Valencian Government (CIT)
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
* MA  02110-1301, USA.
* 
*/

package es.prodevelop.gvsig.mobile.gpe.fmap;

import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Random;

import org.gvsig.gpe.GPERegister;
import org.gvsig.gpe.exceptions.GPEParserCreationException;
import org.gvsig.gpe.exceptions.GPEParserRegisterException;
import org.gvsig.gpe.parser.GPEContentHandler;
import org.gvsig.gpe.parser.GPEErrorHandler;
import org.gvsig.gpe.parser.GPEParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.hardcode.gdbms.engine.data.driver.DriverException;

import es.prodevelop.gvsig.mobile.fmap.MapContext;
import es.prodevelop.gvsig.mobile.fmap.core.FPoint2D;
import es.prodevelop.gvsig.mobile.fmap.core.FPolygon2D;
import es.prodevelop.gvsig.mobile.fmap.core.FPolyline2D;
import es.prodevelop.gvsig.mobile.fmap.core.FShape;
import es.prodevelop.gvsig.mobile.fmap.core.IGeometry;
import es.prodevelop.gvsig.mobile.fmap.core.ShapeFactory;
import es.prodevelop.gvsig.mobile.fmap.driver.BoundedShapes;
import es.prodevelop.gvsig.mobile.fmap.driver.FMapDriverException;
import es.prodevelop.gvsig.mobile.fmap.driver.VectorialFileDriver;
import es.prodevelop.gvsig.mobile.fmap.proj.CRSFactory;
import es.prodevelop.gvsig.mobile.fmap.proj.IProjection;
import es.prodevelop.gvsig.mobile.fmap.util.Utils;
import es.prodevelop.gvsig.mobile.gpe.handlers.FmapContentHandler;
import es.prodevelop.gvsig.mobile.gpe.handlers.FmapErrorHandler;
import es.prodevelop.gvsig.mobile.gpe.handlers.FmapHandlerFactory;
import es.prodevelop.gvsig.mobile.gpe.layer.FlyrGpeVect;
import es.prodevelop.gvsig.mobile.gpe.model.GPEFeature;
import es.prodevelop.gvsig.mobile.gpe.model.GPEGeometry;


/**
 * Implements VectorialFileDriver to create a new 		
 * vectorial driver in order of parse gpe compliant 	
 * files.												
 *														 
 * <br>2008 PRODEVELOP SL</br> 
 * @author Carlos Snchez Perin <br>csanchez@prodevelop.es</br>
 */
public class VectorialGpeDriver implements VectorialFileDriver, BoundedShapes {

	//----------------
	//GLOBAL VARIABLES
	//----------------
	private MapContext mapContext = null;
	private IProjection proj = null;
	private File file = null;
	private GPEParser parser = null;
	private FmapContentHandler contentHandler = null;
	private FmapErrorHandler errorHandler = null;
	private FileInputStream fin;
	private Rectangle2D extent;
	private static Logger logger = LoggerFactory.getLogger(VectorialGpeDriver.class.getName());
		
	//The attributes position
	private LinkedHashMap fieldPosition = null; 
	private HashMap fields = null;
	//The data
	private HashMap features = null;
	private GPEFeature parentShape; 
	
	//-----------
	//CONSTRUCTOR
	//-----------
	public VectorialGpeDriver(MapContext mapControl){
		//set the Context of the map.
		setMapContext(mapControl);
		//get the wanted projection by the user and supported in gvSIG mobile. 
		setProjection((IProjection)mapControl.getProjection());
		//creates the data structures...
		setFields(null);
		setFieldPosition(null);
		setFeatures(null);
	}
	
	/**
	 * Reads the registered parsers in order to check the file formats libGPE 
	 * can parse.
	 * 
	 * @throws GPEParserRegisterException.
	 */
	public void setUp() throws GPEParserRegisterException{
		//---------------
		//VERY IMPORTANT: Fisrts of all we have to register the parser
		//---------------
		//LA SIGUIENTE LINEA PARA ESTABLECER CUALQUIER PARSER.
		GPERegister.addGpeParser("GML v3.1.2 with the SFP","This parser parses GML 3.1.2 SFP",org.gvsig.gpe.gml.parser.GPEGmlSFP0Parser.class);		
		GPERegister.addGpeParser("KML v2.1","This parser parses KML 2.1",org.gvsig.gpe.kml.parser.GPEKmlv21Parser.class);
		GPERegister.addGpeParser("GPX v1.0", "This parser parses GPX 1.0", org.gvsig.gpe.gpx.parser.GPEGpxv11Parser.class);
		//LO TENDR QUE HACER LEYENDO UN FICHERO DE TEXTO CON LOS PARSER'S DISPONIBLES....

		logger.debug(Utils.time() + "INFO: GPE Parsers registered");	
	}

	/**
	 * Set the Extent by a Rectangle2D.
	 */
	public void setExtent(Rectangle2D extent) {
		if (this.extent==null)
			this.extent = extent;
		else
			this.extent.add(extent);		
	}
	
	/**
	 * Set the ErrorHandler.
	 */
	private void setErrorHandler(FmapErrorHandler errorHandler) {			
		this.errorHandler=errorHandler;	
	}

	/**
	 * Get a list of parsed layers.
	 * 
	 * @return Layers	:	ArrayList of Layers (OJO!!! FLyrGpeVectorial).
	 */
	public ArrayList getLayers(){
		contentHandler = (FmapContentHandler) parser.getContentHandler();
		return  contentHandler.getLayers();
	}
	
	/**
	 * Get the ContentHandler, if it don't exists creates a new one.
	 * 
	 * @return contentHandler	:	GPEContentHandler.
	 */
	public GPEContentHandler getContenHandler() {
		if (contentHandler == null){
			contentHandler = new FmapContentHandler(getMapContext(),getErrorHandler());
		}
		return contentHandler;
	}

	/**
	 * Get the ErrorHandler, if it don't exists creates a new one.
	 * 
	 * @return errorHandler	:	GPEErrorHandler.
	 */
	public GPEErrorHandler getErrorHandler() {
		if (errorHandler == null){
			errorHandler = new FmapErrorHandler();
		}
		return errorHandler;
	}
	
	/**
	 * Return true if accepted, in this case only GPX, GML, XML or KML is accepted.
	 */
	public boolean accept(File f) {
		return (f.getName().toUpperCase().endsWith("GML"))
		||(f.getName().toUpperCase().endsWith("XML"))
		||(f.getName().toUpperCase().endsWith("KML"))
		||(f.getName().toUpperCase().endsWith("KMZ"))
		||(f.getName().toUpperCase().endsWith("GPX"));
	}
	
	/**
	 * Close the gpe file.
	 */
	public void close() throws IOException {
		try {
			fin.close();
		} catch (IOException e1) {
			logger.error("While closing file input stream: " + e1.getMessage());
		}
	}
	

	
	/**
	 * Starts libGPE to parse the file.
	 */
	public void initialize(){
		//CREATE THE HANDLERS...
		setErrorHandler(FmapHandlerFactory.createErrorHandler());
		setContentHandler(FmapHandlerFactory.createContentHandler(getMapContext(),getErrorHandler()));
		
		//SET UP THE GPE PARSER'S REGISTER...
		try {
			setUp(); 
			logger.debug(Utils.time() + "INFO: Terminado Registro de parser's GPE.");
		} catch (GPEParserRegisterException e1) {
			/* PUT USER WARNING! */
			e1.printStackTrace();	
			logger.debug(Utils.time() + "ERROR: No se ha podido registrar los parser's de GPE.");
		}			
		//CHECK IF IT'S PARSEABLE FILE 
		File file = getFile();
		logger.debug(Utils.time() + "INFO: Abriendo Fichero: "+ file.getName());
		//IF THERE ARE A CORRECT PARSER TO THE FILE FORMAT, START THE PARSING PROCESS
		//if(GPEFactory.accept(file.toURI())==true){
		if(GPERegister.accept(file.toURI())==true){
			logger.debug(Utils.time() + "INFO: Existe parser registrado para el fomato ");
			
			//CREATES A NEW PARSER AND SET IT AS DEFAULT DRIVER'S PARSER.
			/********************VIEJO REGISTRO********************/
			try {
				setParser((GPERegister.createParser()));
				logger.debug(Utils.time() + "INFO: Creado el parser");
			} catch (GPEParserCreationException e) {
				e.printStackTrace();
				logger.error(Utils.time() + "ERROR: Error Creando el parser");
			}    
			/****************** NUEVO REGISTRO ********************
			try {
				setParser(GPEFactory.createParser(file.toURI()));
			} catch (GPEParserCreationException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}******************************************************/
			
			//CALL TO LIBGPE PARSER TO GET THE PARSED FEATURES THROUGH THE CONTENT HANDLER
			getParser().parse(getContenHandler(),getErrorHandler(), file.toURI());
			logger.debug(Utils.time() +"File "+file.getPath()+" Readed by GPE Parser");
			
			//((FmapErrorHandler)this.getErrorHandler()).showErrors();
			
			//Once parsed the file set the parsed data on the MapContext Layer 
			//The parsed data comes from the ContentHandler
			FlyrGpeVect gpelyr = (FlyrGpeVect) ((FmapContentHandler) getContenHandler()).getParentLayer();
			try {
				setExtent((gpelyr.getDriver()).getFullExtent());
				setFields((gpelyr.getDriver()).getFields());						
				setFieldPosition((gpelyr.getDriver()).getFieldPosition());
				setFeatures((gpelyr.getDriver()).getFeatures());		
			} catch (IOException e) {
				logger.error(Utils.time() +"Error copying data form ContentHandler to Layer by GPE Parser");
				logger.error(e.toString());
				gpelyr=null;
			} catch (NullPointerException ne){
				logger.error(Utils.time()+"Fichero la capa vectorial GPE es nula o no se pudo parsear");
				logger.error(ne.toString());
				gpelyr=null;
			}
		}else
			logger.debug(Utils.time() + "No existe un parser registrado para el fomato seleccionado");	
	}

	/**
	 * Open the GPE file.
	 * 
	 * @param	File to be opened
	 * @throws 	IOException
	 */
	public void open(File f) throws IOException {
		file=f;
		fin = new FileInputStream(file);
	}
	
	/**
	 * Gets the extent from the contentHandler.
	 * 
	 * @return Rectangle2D.
	 * @throws IOException.
	 */
	public Rectangle2D getFullExtent() throws IOException {
		if (extent == null){
			//CALCULE THE NEW EXTENT
			int max = features.size();
			if (max>0){
				Random generator = new Random();
				for (int i=0;i<3;i++){
					int pos = generator.nextInt(max);
					GPEFeature feat = (GPEFeature) features.get(new Integer(pos));
					if (!feat.isComplex()){
						if(extent==null)	
							extent=feat.getGeometry().getShapeBounds();
						else
							extent.add(feat.getGeometry().getShapeBounds());
					}
				}
			}
		}	
		return extent;
	}
	
	/**
	 * Return a geometry by index.
	 *
	 * @param index : number of the geometry
	 * @return IGeometry
	 * @throws IOException
	 */
	public IGeometry getShape(int index){
		IGeometry geom = null;
		GPEFeature shape = (GPEFeature) getFeatures().get(new Integer(index));
		geom = shape.getGeometry().getReprojectedGeometry();
		return geom.cloneGeometry();
	}
	
	/**
	 * Return true if the shape have a 
	 * multiple geometry.
	 * 
	 * @param index : number of the geometry
	 * @return boolean
	 */
	public boolean isMultiShape(int index){
		GPEFeature shape = (GPEFeature) getFeatures().get(new Integer(index));
		if (shape.getGeometry().isMultiGeometry()){
			this.parentShape=shape;
			return true;
		}
		return false;
	}
	
	/**
	 * Return a sub geometry by index.
	 *
	 * @param index : number of the geometry
	 * @return IGeometry
	 */
	public IGeometry getSubShape(int index){
		IGeometry geom = ((GPEGeometry)this.parentShape.getGeometry()).getGeometryAt(index).getGeometry();
		return geom.cloneGeometry();
	}
	
	/**
	 * Return the number of sub geometries.
	 *
	 * @return Number of SubShapes
	 */
	public int getNumSubShapes(){
		return ((GPEGeometry)this.parentShape.getGeometry()).getGeometriesSize();
	}
	
	/**
	 * Return an extent by index of a geometry.
	 *
	 * @param index : number of the geometry
	 * @return Rectangle2D
	 */
	public Rectangle2D getShapeBounds(int index){
		return ((GPEFeature) getFeatures().get(new Integer(index))).getGeometry().getShapeBounds();
	}

	/**
	 * Get the number of geometries to draw.
	 * 
	 * @return numGeometries	:	int
	 */
	public int getShapeCount() throws IOException {
		return features.size();
	}

	/**
	 * Get the type of the geometries to draw 
	 * in this layer.
	 * 
	 * @return shapeType	:	int
	 */
	public int getShapeType() {
		return FShape.MULTI;
	}

	/**
	 * If is writable layer return true.
	 * 
	 * @return boolean
	 */
	public boolean isWritable() {
		return true;
	}
	
	/**
	 * Reload the layer.
	 */
	public void reload() throws IOException, DriverException {
		open(file);
		initialize();
	}

	/**
	 * Return the Name of the Driver.
	 * 
	 * @return String
	 */
	public String getName() {
		return "GPE Vectorial driver";
	}

	/**
	 * Checks the Feature geometries to find if one 
	 * intersects with the point pressed and answer in 
	 * an array which of they intersects.
	 * 
	 * @param	rect		:	Rectangle2D
	 * @param	max_count	:	int
	 * @return  features_intersected[] 	: 	int[]
	 */
	public int[] getIntersectingIndices(Rectangle2D rect, int max_count)
			throws IOException {
		ArrayList pointIntersect = new ArrayList();
		ArrayList lineIntersect = new ArrayList();
		ArrayList polygonIntersect = new ArrayList();
		
		for (int i=0; i<getNumFeatures(); i++) {
			if (Utils.rectanglesItersect(getShapeBounds(i), rect)) {
				IGeometry ig = getShape(i);
				//TODO Have to order points first than lines because 
				//is more important show the point data when it intersects in GPX
				//GPX compose points into a line to show track as a geometry.
				if (ig.intersects(rect)) {	
					if (ig.getInternalShape() instanceof FPoint2D){
						pointIntersect.add(new Integer(i));
						logger.debug("Ha intersectado un Punto");
					}else if(ig.getInternalShape() instanceof FPolyline2D){
						lineIntersect.add(new Integer(i));
						logger.debug("Ha intersectado una Linea");
					}else if(ig.getInternalShape() instanceof FPolygon2D){
						polygonIntersect.add(new Integer(i));
						logger.debug("Ha intersectado un Poligono");
					}
				}
			}
		}
		int numPoints = pointIntersect.size();
		int numLines = lineIntersect.size();
		int numPolygons = polygonIntersect.size();
		
		if ((numPoints == 0)&&(numLines == 0)&&(numPolygons==0)) return null;

		int[] resp = new int[numPoints+numLines+numPolygons];
		int cont=0;
		for (int i=0; i<numPolygons; i++){
			resp[cont] = ((Integer) polygonIntersect.get(i)).intValue();
			cont++;
		}
		for (int j=0; j<numLines; j++){
			resp[cont] = ((Integer) lineIntersect.get(j)).intValue();
			cont++;
		}
		for (int w=0; w<numPoints; w++){
			resp[cont] = ((Integer) pointIntersect.get(w)).intValue();
			cont++;
		}
		return resp;
	}

	/**
	 * Get the geometry from a feature and get the SRS
	 * (SPATIAL REFERENCE SYSTEM) from it.
	 * 
	 * @param	feature		:	GPEFeature
	 * @param	projection	:	IProjection
	 * @return 	geometry	:	IGeometry
	 */
	public IGeometry getGeometry(GPEFeature feature,  IProjection projection){
		GPEGeometry gpeGeometry = feature.getGeometry();
		if (gpeGeometry != null){
			IProjection geomProj = null;
			if (gpeGeometry.getSrs() != null){
				try{
					geomProj = CRSFactory.getCRS(gpeGeometry.getSrs());
				}catch(Exception e){
					logger.error("The CRS Factory has an error while where seeking the SRS from the geometry.");
					logger.error(e.toString());
				}				
			}
			if ((geomProj == null) || (geomProj.getAbrev().compareTo(projection.getAbrev()) == 0)){
				if (gpeGeometry.isMultiGeometry()){
					return ShapeFactory.createGeometry((FShape) ComposeShape(gpeGeometry).getInternalShape());
				}
				else{
					return gpeGeometry.getGeometry();
				}
			}else{
					FShape shape = (FShape)gpeGeometry.getGeometry().getInternalShape();
					// Reprojection the shape
					//if (coordTrans!= null)
					//	  shape.reProject(coordTrans);
					return ShapeFactory.createGeometry(shape);
			}
		}
		return null;
	}
	
	/**
	 * Compose a new Shape from the multishape include into a multigeometry
	 * Recursive algorithm.
	 *  
	 * @param GPEGeometry Geometry model from GPE
	 */
	private IGeometry ComposeShape(GPEGeometry gpeGeometry) {
		ArrayList list= new ArrayList();
		for (int i=0;i<gpeGeometry.getGeometriesSize();i++){
			GPEGeometry geom = gpeGeometry.getGeometryAt(i);
			if (geom.isMultiGeometry()){
				return ComposeShape(geom);
			}
			else{
				IGeometry geometry = geom.getGeometry();
				list.add(geometry.getInternalShape());	
			}
		}
		if(gpeGeometry.getGeometry()!=null)
			list.add(gpeGeometry.getGeometry().getInternalShape());
		if(list.get(0) instanceof FPolygon2D){
			return ShapeFactory.createPolygon2D(list);
		}
		else if(list.get(0) instanceof FPolyline2D){
			return ShapeFactory.createPolyline2D(list);
		}
		else{
			return null;
		}
	}

	/**
	 * Return Data File the same in GPE as the parsered one.
	 * 
	 * @return File
	 */
	public File getDataFile(File f) {
		return file;
	}

	/**
	 * Return the File parsered.
	 * 
	 * @return File
	 */
	public File getFile() {
		return file;
	}	
	
	/**
	 * Return the parsered File absolute path.
	 * 
	 * @return String
	 */
	public String getFilePath() {
		return file.getAbsolutePath();
	}
	
	/**
	 * Return the Shape Type.
	 * 
	 * @return integer Type
	 */
	public int getShapeType(int index) {
		int auxType = 0;		
		return auxType;
	}
	
	/**
	 * Return Data Source Name.
	 * 
	 * @return String Name
	 */
	public String getDataDriverName() {
		return VectorialGpeAdapter.GPE_FILE_DATASOURCE;
	}
	
	/**
	 * Return Number of Features.
	 * 
	 * @return integer features.
	 */
	public int getNumFeatures() {
		return features.size();
	}
	
	/**
	 * Return List of attributes.
	 * 
	 * @return Array List field names
	 */
	public ArrayList getAttributes() {
		ArrayList attrib = new ArrayList(); 
		for (int i=0;i<getFeatures().size();i++){
			GPEFeature element = (GPEFeature) getFeatures().get(new Integer(i));
			attrib.add(element);	
		}
		return attrib;
	}
	
	/**
	 * Return the Table of fields content.
	 * 
	 * @return HashMap of fields
	 */
	public HashMap getFields() {
		return fields;
	}
	
	/**
	 * Return the Index Table of fields position.
	 * 
	 * @return HashMap of positions
	 */
	public LinkedHashMap getPosition() {
		return fieldPosition;
	}
	
	/**
	 * Return the Table of Features.(Contains Geodata and Data)
	 * 
	 * @return HashMap of features
	 */
	public HashMap getFeatures() {
		return features;
	}
	
	/**
	 * Return Actual Parsered File Projection.
	 * 
	 * @return projection
	 */
	public IProjection getProjection() {
		return proj;
	}
	
	/**
	 * Return the MapContext.
	 * 
	 * @return mapContext
	 */
	public MapContext getMapContext() {
		return mapContext;
	}
	
	/**
	 * Return ordered indexes of fields position.
	 * 
	 * @return HasMap
	 */
	public LinkedHashMap getFieldPosition() {
		return fieldPosition;
	}
	
	/**
	 * Return the Parser from GPE.
	 * 
	 * @return The GPEParser
	 */
	public GPEParser getParser() {
		return parser;
	}
	
	//-----------------------------------------
	/******************************************
	 * Setters(Object).
	 * ----------------
	 * Set the value of	an Object...
	 * 
	 * @param Object
	 ******************************************/
	
	//-----------------------------------------
	public void setContentHandler(FmapContentHandler contentHandler) {
		this.contentHandler = contentHandler;
	}
	//------------------------------------------
	public void setParser(GPEParser parser) {
		this.parser = parser;
	}
	//------------------------------------------	
	public void setProjection(IProjection proj) {
		this.proj = proj;
	}
	//------------------------------------------
	public void setMapContext(MapContext mapContext) {
		this.mapContext = mapContext;
	}
	//------------------------------------------
	public void setFieldPosition(LinkedHashMap fieldPosition) {
		if (fieldPosition==null)
			this.fieldPosition = new LinkedHashMap();
		else{
			this.fieldPosition = fieldPosition;
		}
	}
	//------------------------------------------
	public void setFields(HashMap fields) {
		if (fields==null){
			this.fields = new HashMap();
		}else{
			this.fields = fields;
		}
	}
	//------------------------------------------
	public void setFeatures(HashMap features) {
		if (features==null)
			this.features = new HashMap();
		else
			this.features = features;
	}

}
